@@ -33,8 +33,10 @@ extern "C" {
 #include <stdint.h>
 #include <inttypes.h>
 #include <stdlib.h>
-#include <stdarg.h>
 #include <unistd.h>
+#include <math.h>
+
+#include <spa/utils/string.h>
 
 typedef struct pa_core pa_core;
 
@@ -45,10 +47,12 @@ typedef void (*pa_free_cb_t)(void *p);
 #define PA_LIKELY(x) (__builtin_expect(!!(x),1))
 #define PA_UNLIKELY(x) (__builtin_expect(!!(x),0))
 #define PA_PRINTF_FUNC(fmt, arg1) __attribute__((format(printf, fmt, arg1)))
+#define PA_UNUSED __attribute__ ((unused))
 #else
 #define PA_LIKELY(x) (x)
 #define PA_UNLIKELY(x) (x)
 #define PA_PRINTF_FUNC(fmt, arg1)
+#define PA_UNUSED
 #endif
 
 #define PA_MIN(a,b)                    \
@@ -94,7 +98,7 @@ typedef enum pa_available {
 	PA_AVAILABLE_YES = 2,
 } pa_available_t;
 
-#define PA_RATE_MAX (48000U*8U)
+#define PA_RATE_MAX (48000U*16U)
 
 typedef enum pa_sample_format {
 	PA_SAMPLE_U8,		/**< Unsigned 8 Bit PCM */
@@ -281,7 +285,44 @@ static inline PA_PRINTF_FUNC(5, 6) void pa_log_level_meta(enum pa_log_level leve
 #define pa_strnull(s)	((s) ? (s) : "null")
 #define pa_startswith(s,pfx)	(strstr(s, pfx) == s)
 
-#define pa_snprintf	snprintf
+PA_PRINTF_FUNC(3, 0)
+static inline size_t pa_vsnprintf(char *str, size_t size, const char *format, va_list ap)
+{
+	int ret;
+
+	pa_assert(str);
+	pa_assert(size > 0);
+	pa_assert(format);
+
+	ret = vsnprintf(str, size, format, ap);
+
+	str[size-1] = 0;
+
+	if (ret < 0)
+		return strlen(str);
+
+	if ((size_t) ret > size-1)
+		return size-1;
+
+	return (size_t) ret;
+}
+
+PA_PRINTF_FUNC(3, 4)
+static inline size_t pa_snprintf(char *str, size_t size, const char *format, ...)
+{
+	size_t ret;
+	va_list ap;
+
+	pa_assert(str);
+	pa_assert(size > 0);
+	pa_assert(format);
+
+	va_start(ap, format);
+	ret = pa_vsnprintf(str, size, format, ap);
+	va_end(ap);
+
+	return ret;
+}
 
 #define pa_xstrdup(s)		((s) != NULL ? strdup(s) : NULL)
 #define pa_xstrndup(s,n)	((s) != NULL ? strndup(s,n) : NULL)
@@ -309,6 +350,48 @@ static inline void pa_xstrfreev(char **a) {
     pa_xfreev((void**)a);
 }
 
+typedef struct {
+	size_t size;
+	char *ptr;
+	FILE *f;
+} pa_strbuf;
+
+static inline pa_strbuf *pa_strbuf_new(void)
+{
+	pa_strbuf *s = pa_xnew0(pa_strbuf,1);
+	s->f = open_memstream(&s->ptr, &s->size);
+	return s;
+}
+
+static PA_PRINTF_FUNC(2,3) inline size_t pa_strbuf_printf(pa_strbuf *sb, const char *format, ...)
+{
+	int ret;
+	va_list args;
+	va_start(args, format);
+	ret = vfprintf(sb->f, format, args);
+	va_end(args);
+	return ret > 0 ? ret : 0;
+}
+
+static inline void pa_strbuf_puts(pa_strbuf *sb, const char *t)
+{
+	fputs(t, sb->f);
+}
+
+static inline bool pa_strbuf_isempty(pa_strbuf *sb)
+{
+	fflush(sb->f);
+	return sb->size == 0;
+}
+
+static inline char *pa_strbuf_to_string_free(pa_strbuf *sb)
+{
+	char *ptr;
+	fclose(sb->f);
+	ptr = sb->ptr;
+	free(sb);
+	return ptr;
+}
 
 #define pa_cstrerror	strerror
 
@@ -509,25 +592,35 @@ static inline char *pa_strip(char *s)
 
 static inline int pa_atod(const char *s, double *ret_d)
 {
-	char *x;
-	*ret_d = strtod(s, &x);
-	return 0;
+	if (spa_atod(s, ret_d) && !isnan(*ret_d))
+		return 0;
+	errno = EINVAL;
+	return -1;
 }
 static inline int pa_atoi(const char *s, int32_t *ret_i)
 {
-	*ret_i = (int32_t) atoi(s);
-	return 0;
+	if (spa_atoi32(s, ret_i, 0))
+		return 0;
+	errno = EINVAL;
+	return -1;
 }
 static inline int pa_atou(const char *s, uint32_t *ret_u)
 {
-	*ret_u = (uint32_t) atoi(s);
-	return 0;
+	if (spa_atou32(s, ret_u, 0))
+		return 0;
+	errno = EINVAL;
+	return -1;
 }
 static inline int pa_atol(const char *s, long *ret_l)
 {
-	char *x;
-	*ret_l = strtol(s, &x, 0);
-	return 0;
+	int64_t res;
+	if (spa_atoi64(s, &res, 0)) {
+		*ret_l = res;
+		if (*ret_l == res)
+			return 0;
+	}
+	errno = EINVAL;
+	return -1;
 }
 
 static inline int pa_parse_boolean(const char *v)
