@@ -1,27 +1,7 @@
-/* Spa HSP/HFP native backend
- *
- * Copyright © 2018 Wim Taymans
- * Copyright © 2021 Collabora
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* Spa HSP/HFP native backend */
+/* SPDX-FileCopyrightText: Copyright © 2018 Wim Taymans */
+/* SPDX-FileCopyrightText: Copyright © 2021 Collabora */
+/* SPDX-License-Identifier: MIT */
 
 #include <errno.h>
 #include <unistd.h>
@@ -36,6 +16,8 @@
 
 #include <dbus/dbus.h>
 
+#include <spa/debug/mem.h>
+#include <spa/debug/log.h>
 #include <spa/support/log.h>
 #include <spa/support/loop.h>
 #include <spa/support/dbus.h>
@@ -58,6 +40,7 @@ static struct spa_log_topic log_topic = SPA_LOG_TOPIC(0, "spa.bluez5.native");
 #undef SPA_LOG_TOPIC_DEFAULT
 #define SPA_LOG_TOPIC_DEFAULT &log_topic
 
+#define PROP_KEY_ROLES "bluez5.roles"
 #define PROP_KEY_HEADSET_ROLES "bluez5.headset-roles"
 
 #define HFP_CODEC_SWITCH_INITIAL_TIMEOUT_MSEC 5000
@@ -129,6 +112,8 @@ struct impl {
 struct transport_data {
 	struct rfcomm *rfcomm;
 	struct spa_source sco;
+	int err;
+	bool requesting;
 };
 
 enum hfp_hf_state {
@@ -779,6 +764,12 @@ static bool rfcomm_hfp_ag(struct rfcomm *rfcomm, char* buf)
 	int xapl_product;
 	int xapl_features;
 
+	spa_debug_log_mem(backend->log, SPA_LOG_LEVEL_DEBUG, 2, buf, strlen(buf));
+
+	/* Some devices send initial \n: be permissive */
+	while (*buf == '\n')
+		++buf;
+
 	if (sscanf(buf, "AT+BRSF=%u", &features) == 1) {
 		unsigned int ag_features = SPA_BT_HFP_AG_FEATURE_NONE;
 
@@ -880,6 +871,9 @@ static bool rfcomm_hfp_ag(struct rfcomm *rfcomm, char* buf)
 				rfcomm_emit_volume_changed(rfcomm, -1, SPA_BT_VOLUME_INVALID);
 			}
 		}
+	} else if (spa_streq(buf, "\r")) {
+		/* No commands, reply OK (ITU-T Rec. V.250 Sec. 5.2.1 & 5.6) */
+		rfcomm_send_reply(rfcomm, "OK");
 	} else if (!rfcomm->slc_configured) {
 		spa_log_warn(backend->log, "RFCOMM receive command before SLC completed: %s", buf);
 		rfcomm_send_error(rfcomm, CMEE_AG_FAILURE);
@@ -929,6 +923,15 @@ static bool rfcomm_hfp_ag(struct rfcomm *rfcomm, char* buf)
 		rfcomm_send_reply(rfcomm, "OK");
 		if (was_switching_codec)
 			spa_bt_device_emit_codec_switched(rfcomm->device, 0);
+	} else if (spa_strstartswith(buf, "AT+BCC")) {
+		if (!rfcomm->codec_negotiation_supported)
+			return false;
+
+		rfcomm_send_reply(rfcomm, "OK");
+		rfcomm_send_reply(rfcomm, "+BCS: %u", rfcomm->codec);
+		rfcomm->hfp_ag_switching_codec = true;
+		rfcomm->hfp_ag_initial_codec_setup = HFP_AG_INITIAL_CODEC_SETUP_NONE;
+		codec_switch_start_timer(rfcomm, HFP_CODEC_SWITCH_TIMEOUT_MSEC);
 	} else if (spa_strstartswith(buf, "AT+BIA=")) {
 		/* retrieve indicators activation
 		 * form: AT+BIA=[indrep1],[indrep2],[indrepx] */
@@ -1361,7 +1364,7 @@ static int sco_create_socket(struct impl *backend, struct spa_bt_adapter *adapte
 	bdaddr_t src;
 	int sock = -1;
 
-	sock = socket(PF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_SCO);
+	sock = socket(PF_BLUETOOTH, SOCK_SEQPACKET | SOCK_NONBLOCK, BTPROTO_SCO);
 	if (sock < 0) {
 		spa_log_error(backend->log, "socket(SEQPACKET, SCO) %s", strerror(errno));
 		goto fail;
@@ -1414,6 +1417,8 @@ static int sco_do_connect(struct spa_bt_transport *t)
 	spa_log_debug(backend->log, "transport %p: enter sco_do_connect, codec=%u",
 			t, t->codec);
 
+	td->err = -EIO;
+
 	if (d->adapter == NULL)
 		return -EIO;
 
@@ -1456,6 +1461,8 @@ again:
 		goto fail_close;
 	}
 
+	td->err = -EINPROGRESS;
+
 	return sock;
 
 fail_close:
@@ -1465,37 +1472,76 @@ fail_close:
 
 static int rfcomm_ag_sync_volume(struct rfcomm *rfcomm, bool later);
 
-static void wait_for_socket(int fd)
+static void sco_ready(struct spa_bt_transport *t)
 {
-	struct pollfd fds[1];
-	const int timeout_ms = 500;
-
-	fds[0].fd = fd;
-	fds[0].events = POLLIN | POLLERR | POLLHUP;
-	poll(fds, 1, timeout_ms);
-}
-
-static int sco_acquire_cb(void *data, bool optional)
-{
-	struct spa_bt_transport *t = data;
-	struct transport_data *td = t->user_data;
 	struct impl *backend = SPA_CONTAINER_OF(t->backend, struct impl, this);
-	int sock;
+	struct transport_data *td = t->user_data;
+	struct sco_options sco_opt;
 	socklen_t len;
+	int err;
 
-	spa_log_debug(backend->log, "transport %p: enter sco_acquire_cb", t);
+	spa_log_debug(backend->log, "transport %p: ready", t);
 
-	if (optional || t->fd > 0)
-		sock = t->fd;
-	else
-		sock = sco_do_connect(t);
+	/* Read socket error status */
+	if (t->fd >= 0) {
+		if (td->err == -EINPROGRESS) {
+			len = sizeof(err);
+			memset(&err, 0, len);
+			if (getsockopt(t->fd, SOL_SOCKET, SO_ERROR, &err, &len) < 0)
+				td->err = -errno;
+			else
+				td->err = -err;
+		}
+	} else {
+		td->err = -EIO;
+	}
 
-	if (sock < 0)
-		goto fail;
+	if (!td->requesting)
+		return;
 
-#ifdef HAVE_BLUEZ_5_BACKEND_HFP_NATIVE
-	rfcomm_hfp_ag_set_cind(td->rfcomm, true);
-#endif
+	td->requesting = false;
+
+	if (td->err)
+		goto done;
+
+	/* XXX: The MTU as currently reported by kernel (6.2) here is not a valid packet size,
+	 * XXX: for USB adapters, see sco-io.
+	 */
+	len = sizeof(sco_opt);
+	memset(&sco_opt, 0, len);
+	if (getsockopt(t->fd, SOL_SCO, SCO_OPTIONS, &sco_opt, &len) < 0) {
+		spa_log_warn(backend->log, "getsockopt(SCO_OPTIONS) failed, using defaults");
+		t->read_mtu = 48;
+		t->write_mtu = 48;
+	} else {
+		spa_log_debug(backend->log, "autodetected mtu = %u", sco_opt.mtu);
+		t->read_mtu = sco_opt.mtu;
+		t->write_mtu = sco_opt.mtu;
+	}
+
+	/* Clear nonblocking flag we set for connect() */
+	err = fcntl(t->fd, F_GETFL, O_NONBLOCK);
+	if (err < 0) {
+		td->err = -errno;
+		goto done;
+	}
+	err &= ~O_NONBLOCK;
+	err = fcntl(t->fd, F_SETFL, O_NONBLOCK, err);
+	if (err < 0) {
+		td->err = -errno;
+		goto done;
+	}
+
+done:
+	if (td->err) {
+		spa_log_debug(backend->log, "transport %p: acquire failed: %s (%d)",
+				t, strerror(-td->err), td->err);
+		spa_bt_transport_set_state(t, SPA_BT_TRANSPORT_STATE_ERROR);
+		return;
+	}
+
+	spa_log_debug(backend->log, "transport %p: acquire complete, read_mtu=%u, write_mtu=%u",
+			t, t->read_mtu, t->write_mtu);
 
 	/*
 	 * Send RFCOMM volume after connection is ready, and also after
@@ -1512,55 +1558,65 @@ static int sco_acquire_cb(void *data, bool optional)
 	 * volume buttons, which is updated by an +VGS event only after
 	 * sufficient time is elapsed from the connection.
 	 */
-	wait_for_socket(sock);
 	rfcomm_ag_sync_volume(td->rfcomm, false);
 	rfcomm_ag_sync_volume(td->rfcomm, true);
 
+	spa_bt_transport_set_state(t, SPA_BT_TRANSPORT_STATE_ACTIVE);
+}
+
+static void sco_start_source(struct spa_bt_transport *t);
+
+static int sco_acquire_cb(void *data, bool optional)
+{
+	struct spa_bt_transport *t = data;
+	struct transport_data *td = t->user_data;
+	struct impl *backend = SPA_CONTAINER_OF(t->backend, struct impl, this);
+	int sock;
+
+	spa_log_debug(backend->log, "transport %p: enter sco_acquire_cb", t);
+
+	if (optional || t->fd > 0)
+		sock = t->fd;
+	else
+		sock = sco_do_connect(t);
+
+	if (sock < 0)
+		goto fail;
+
+#ifdef HAVE_BLUEZ_5_BACKEND_HFP_NATIVE
+	rfcomm_hfp_ag_set_cind(td->rfcomm, true);
+#endif
+
 	t->fd = sock;
 
-	/* Fallback value */
-	t->read_mtu = 48;
-	t->write_mtu = 48;
+	td->requesting = true;
 
-	if (true) {
-		struct sco_options sco_opt;
+	sco_start_source(t);
 
-		len = sizeof(sco_opt);
-		memset(&sco_opt, 0, len);
-
-		if (getsockopt(sock, SOL_SCO, SCO_OPTIONS, &sco_opt, &len) < 0)
-			spa_log_warn(backend->log, "getsockopt(SCO_OPTIONS) failed, loading defaults");
-		else {
-			spa_log_debug(backend->log, "autodetected mtu = %u", sco_opt.mtu);
-			t->read_mtu = sco_opt.mtu;
-			t->write_mtu = sco_opt.mtu;
-		}
-	}
-	spa_log_debug(backend->log, "transport %p: read_mtu=%u, write_mtu=%u", t, t->read_mtu, t->write_mtu);
+	if (td->err != -EINPROGRESS)
+		sco_ready(t);
 
 	return 0;
 
 fail:
+	spa_bt_transport_set_state(t, SPA_BT_TRANSPORT_STATE_ERROR);
 	return -1;
 }
 
-static int sco_release_cb(void *data)
+static int sco_destroy_cb(void *data)
 {
 	struct spa_bt_transport *t = data;
 	struct transport_data *td = t->user_data;
 	struct impl *backend = SPA_CONTAINER_OF(t->backend, struct impl, this);
 
-	spa_log_info(backend->log, "Transport %s released", t->path);
-
-#ifdef HAVE_BLUEZ_5_BACKEND_HFP_NATIVE
-	rfcomm_hfp_ag_set_cind(td->rfcomm, false);
-#endif
-
 	if (t->sco_io) {
 		spa_bt_sco_io_destroy(t->sco_io);
 		t->sco_io = NULL;
 	}
 
+	if (td->sco.loop)
+		spa_loop_remove_source(backend->main_loop, &td->sco);
+
 	if (t->fd > 0) {
 		/* Shutdown and close the socket */
 		shutdown(t->fd, SHUT_RDWR);
@@ -1571,6 +1627,25 @@ static int sco_release_cb(void *data)
 	return 0;
 }
 
+static int sco_release_cb(void *data)
+{
+	struct spa_bt_transport *t = data;
+	struct transport_data *td = t->user_data;
+	struct impl *backend = SPA_CONTAINER_OF(t->backend, struct impl, this);
+
+	spa_log_info(backend->log, "Transport %s released", t->path);
+
+	spa_bt_transport_set_state(t, SPA_BT_TRANSPORT_STATE_IDLE);
+
+#ifdef HAVE_BLUEZ_5_BACKEND_HFP_NATIVE
+	rfcomm_hfp_ag_set_cind(td->rfcomm, false);
+#endif
+
+	sco_destroy_cb(t);
+
+	return 0;
+}
+
 static void sco_event(struct spa_source *source)
 {
 	struct spa_bt_transport *t = source->data;
@@ -1578,15 +1653,44 @@ static void sco_event(struct spa_source *source)
 
 	if (source->rmask & (SPA_IO_HUP | SPA_IO_ERR)) {
 		spa_log_debug(backend->log, "transport %p: error on SCO socket: %s", t, strerror(errno));
+		sco_ready(t);
+		if (source->loop)
+			spa_loop_remove_source(source->loop, source);
 		if (t->fd >= 0) {
-			if (source->loop)
-				spa_loop_remove_source(source->loop, source);
-			shutdown(t->fd, SHUT_RDWR);
-			close (t->fd);
-			t->fd = -1;
 			spa_bt_transport_set_state(t, SPA_BT_TRANSPORT_STATE_IDLE);
+			shutdown(t->fd, SHUT_RDWR);
+			close(t->fd);
+			t->fd = -1;
 		}
 	}
+
+	if (source->rmask & SPA_IO_IN) {
+		SPA_FLAG_UPDATE(source->mask, SPA_IO_IN, false);
+		spa_loop_update_source(backend->main_loop, source);
+		sco_ready(t);
+	}
+}
+
+static void sco_start_source(struct spa_bt_transport *t)
+{
+	struct impl *backend = SPA_CONTAINER_OF(t->backend, struct impl, this);
+	struct transport_data *td = t->user_data;
+
+	if (td->sco.loop)
+		return;
+
+	td->err = -EINPROGRESS;
+
+	/*
+	 * We on purpose wait for POLLIN when connecting (not POLLOUT as usual), to
+	 * indicate ready only after we are sure the device is sending data.
+	 */
+	td->sco.func = sco_event;
+	td->sco.data = t;
+	td->sco.fd = t->fd;
+	td->sco.mask = SPA_IO_HUP | SPA_IO_ERR | SPA_IO_IN;
+	td->sco.rmask = 0;
+	spa_loop_add_source(backend->main_loop, &td->sco);
 }
 
 static void sco_listen_event(struct spa_source *source)
@@ -1598,7 +1702,6 @@ static void sco_listen_event(struct spa_source *source)
 	char local_address[18], remote_address[18];
 	struct rfcomm *rfcomm;
 	struct spa_bt_transport *t = NULL;
-	struct transport_data *td;
 
 	if (source->rmask & (SPA_IO_HUP | SPA_IO_ERR)) {
 		spa_log_error(backend->log, "error listening SCO connection: %s", strerror(errno));
@@ -1681,13 +1784,7 @@ static void sco_listen_event(struct spa_source *source)
 
 	t->fd = sock;
 
-	td = t->user_data;
-	td->sco.func = sco_event;
-	td->sco.data = t;
-	td->sco.fd = sock;
-	td->sco.mask = SPA_IO_HUP | SPA_IO_ERR;
-	td->sco.rmask = 0;
-	spa_loop_add_source(backend->main_loop, &td->sco);
+	sco_start_source(t);
 
 	spa_log_debug(backend->log, "transport %p: audio connected", t);
 
@@ -1823,6 +1920,7 @@ static const struct spa_bt_transport_implementation sco_transport_impl = {
 	.acquire = sco_acquire_cb,
 	.release = sco_release_cb,
 	.set_volume = sco_set_volume_cb,
+	.destroy = sco_destroy_cb,
 };
 
 static struct rfcomm *device_find_rfcomm(struct impl *backend, struct spa_bt_device *device)
@@ -2706,8 +2804,11 @@ static int parse_headset_roles(struct impl *backend, const struct spa_dict *info
 	const char *str;
 	int profiles = SPA_BT_PROFILE_NULL;
 
-	if (info == NULL ||
-	    (str = spa_dict_lookup(info, PROP_KEY_HEADSET_ROLES)) == NULL)
+	if (!info)
+		goto fallback;
+
+	if ((str = spa_dict_lookup(info, PROP_KEY_ROLES)) == NULL &&
+			(str = spa_dict_lookup(info, PROP_KEY_HEADSET_ROLES)) == NULL)
 		goto fallback;
 
 	profiles = spa_bt_profiles_from_json_array(str);
