@@ -1,26 +1,6 @@
-/* Spa Bluez5 Device
- *
- * Copyright © 2018 Wim Taymans
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* Spa Bluez5 Device */
+/* SPDX-FileCopyrightText: Copyright © 2018 Wim Taymans */
+/* SPDX-License-Identifier: MIT */
 
 #include <stddef.h>
 #include <stdio.h>
@@ -1058,13 +1038,16 @@ static void profiles_changed(void *userdata, uint32_t prev_profiles, uint32_t pr
 
 static void set_initial_profile(struct impl *this);
 
-static void device_connected(void *userdata, bool connected) {
+static void device_connected(void *userdata, bool connected)
+{
 	struct impl *this = userdata;
 
 	spa_log_debug(this->log, "connected: %d", connected);
 
-	if (connected ^ (this->profile != DEVICE_PROFILE_OFF))
+	if (connected ^ (this->profile != DEVICE_PROFILE_OFF)) {
+		emit_remove_nodes(this);
 		set_initial_profile(this);
+	}
 }
 
 static const struct spa_bt_device_events bt_dev_events = {
