@@ -1,26 +1,6 @@
-/* Spa Bluez5 Monitor
- *
- * Copyright © 2018 Wim Taymans
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* Spa Bluez5 Monitor */
+/* SPDX-FileCopyrightText: Copyright © 2018 Wim Taymans */
+/* SPDX-License-Identifier: MIT */
 
 #ifndef SPA_BLUEZ5_DEFS_H
 #define SPA_BLUEZ5_DEFS_H
@@ -542,6 +522,8 @@ void spa_bt_device_update_last_bluez_action_time(struct spa_bt_device *device);
 #define spa_bt_device_add_listener(d,listener,events,data)           \
 	spa_hook_list_append(&(d)->listener_list, listener, events, data)
 
+struct spa_bt_iso_io;
+
 struct spa_bt_sco_io;
 
 struct spa_bt_sco_io *spa_bt_sco_io_create(struct spa_loop *data_loop, int fd, uint16_t read_mtu, uint16_t write_mtu);
@@ -559,9 +541,10 @@ int spa_bt_sco_io_write(struct spa_bt_sco_io *io, uint8_t *data, int size);
 #define SPA_BT_VOLUME_A2DP_MAX	127
 
 enum spa_bt_transport_state {
-        SPA_BT_TRANSPORT_STATE_IDLE,
-        SPA_BT_TRANSPORT_STATE_PENDING,
-        SPA_BT_TRANSPORT_STATE_ACTIVE,
+        SPA_BT_TRANSPORT_STATE_ERROR = -1,
+        SPA_BT_TRANSPORT_STATE_IDLE = 0,
+        SPA_BT_TRANSPORT_STATE_PENDING = 1,
+        SPA_BT_TRANSPORT_STATE_ACTIVE = 2,
 };
 
 struct spa_bt_transport_events {
@@ -620,15 +603,22 @@ struct spa_bt_transport {
 	int acquire_refcount;
 	bool acquired;
 	bool keepalive;
+	int error_count;
+	uint64_t last_error_time;
 	int fd;
 	uint16_t read_mtu;
 	uint16_t write_mtu;
-	uint16_t delay;
+	unsigned int delay_us;
+	unsigned int latency_us;
+	uint8_t bap_cig;
+	uint8_t bap_cis;
 
+	struct spa_bt_iso_io *iso_io;
 	struct spa_bt_sco_io *sco_io;
 
 	struct spa_source volume_timer;
 	struct spa_source release_timer;
+	DBusPendingCall *acquire_call;
 
 	struct spa_hook_list listener_list;
 	struct spa_callbacks impl;
