@@ -1,27 +1,7 @@
-/* PipeWire
- *
- * Copyright © 2021 Wim Taymans
- *           © 2021 Arun Raghavan <arun@asymptotic.io>
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* PipeWire */
+/* SPDX-FileCopyrightText: Copyright © 2021 Wim Taymans */
+/* SPDX-FileCopyrightText: Copyright © 2021 Arun Raghavan <arun@asymptotic.io> */
+/* SPDX-License-Identifier: MIT */
 
 #include "config.h"
 
@@ -38,10 +18,10 @@
 #include <sys/stat.h>
 #include <unistd.h>
 
-#include <spa/debug/pod.h>
+#include <spa/debug/types.h>
 #include <spa/param/audio/format-utils.h>
 #include <spa/param/audio/raw.h>
-#include <spa/param/profiler.h>
+#include <spa/param/latency-utils.h>
 #include <spa/pod/builder.h>
 #include <spa/pod/dynamic.h>
 #include <spa/support/plugin.h>
@@ -53,7 +33,8 @@
 #include <spa/support/plugin-loader.h>
 #include <spa/interfaces/audio/aec.h>
 
-#include <pipewire/private.h>
+#include <spa/plugins/audioconvert/wavfile.h>
+
 #include <pipewire/impl.h>
 #include <pipewire/pipewire.h>
 
@@ -65,6 +46,40 @@
  * virtual `echo-cancel-capture` source and `echo-cancel-playback` sink
  * nodes and the associated streams.
  *
+ * The echo-cancel module is mostly used in video or audio conference
+ * applications. When the other participants talk and the audio is going out to
+ * the speakers, the signal will be picked up again by the microphone and sent
+ * back to the other participants (along with your talking), resulting in an
+ * echo. This is annoying because the other participants will hear their own
+ * echo from you.
+ *
+ * Conceptually the echo-canceler is composed of 4 streams:
+ *
+ *\code{.unparsed}
+ * .--------.     .---------.     .--------.     .----------.     .-------.
+ * |  mic   | --> | capture | --> |        | --> |  source  | --> |  app  |
+ * '--------'     '---------'     | echo   |     '----------'     '-------'
+ *                                | cancel |
+ * .--------.     .---------.     |        |     .----------.     .---------.
+ * |  app   | --> |  sink   | --> |        | --> | playback | --> | speaker |
+ * '--------'     '---------'     '--------'     '----------'     '---------'
+ *\endcode
+
+ * - A capture stream that captures audio from a microphone.
+ * - A Sink that takes the signal containing the data that should be canceled
+ *   out from the capture stream. This is where the application (video conference
+ *   application) send the audio to and it contains the signal from the other
+ *   participants that are speaking and that needs to be cancelled out.
+ * - A playback stream that just passes the signal from the Sink to the speaker.
+ *   This is so that you can hear the other participants. It is also the signal
+ *   that gets picked up by the microphone and that eventually needs to be
+ *   removed again.
+ * - A Source that exposes the echo-canceled data captured from the capture
+ *   stream. The data from the sink stream and capture stream are correlated and
+ *   the signal from the sink stream is removed from the capture stream data.
+ *   This data then goes into the application (the conference application) and
+ *   does not contain the echo from the other participants anymore.
+ *
  * ## Module Options
  *
  * Options specific to the behavior of this module
@@ -76,6 +91,8 @@
  * - `library.name = <str>`: the echo cancellation library  Currently supported:
  * `aec/libspa-aec-webrtc`. Leave unset to use the default method (`aec/libspa-aec-webrtc`).
  * - `aec.args = <str>`: arguments to pass to the echo cancellation method
+ * - `monitor.mode`: Instead of making a sink, make a stream that captures from
+ *                   the monitor ports of the default sink.
  *
  * ## General options
  *
@@ -101,6 +118,7 @@
  *      args = {
  *          # library.name  = aec/libspa-aec-webrtc
  *          # node.latency = 1024/48000
+ *          # monitor.mode = false
  *          capture.props = {
  *             node.name = "Echo Cancellation Capture"
  *          }
@@ -121,13 +139,6 @@
  */
 
 /**
- * .--------.     .---------.     .--------.     .----------.     .-------.
- * | source | --> | capture | --> |        | --> |  source  | --> |  app  |
- * '--------'     '---------'     | echo   |     '----------'     '-------'
- *                                | cancel |
- * .--------.     .---------.     |        |     .----------.     .--------.
- * |  app   | --> |  sink   | --> |        | --> | playback | --> |  sink  |
- * '--------'     '---------'     '--------'     '----------'     '--------'
  */
 #define NAME "echo-cancel"
 
@@ -146,19 +157,19 @@ PW_LOG_TOPIC_STATIC(mod_topic, "mod." NAME);
 static const struct spa_dict_item module_props[] = {
 	{ PW_KEY_MODULE_AUTHOR, "Wim Taymans <wim.taymans@gmail.com>" },
 	{ PW_KEY_MODULE_DESCRIPTION, "Echo Cancellation" },
-	{ PW_KEY_MODULE_USAGE, " [ remote.name=<remote> ] "
-				"[ node.latency=<latency as fraction> ] "
-				"[ audio.rate=<sample rate> ] "
-				"[ audio.channels=<number of channels> ] "
-				"[ audio.position=<channel map> ] "
-				"[ buffer.max_size=<max buffer size in ms> ] "
-				"[ buffer.play_delay=<delay as fraction> ] "
-				"[ library.name =<library name> ] "
-				"[ aec.args=<aec arguments> ] "
-				"[ capture.props=<properties> ] "
-				"[ source.props=<properties> ] "
-				"[ sink.props=<properties> ] "
-				"[ playback.props=<properties> ] " },
+	{ PW_KEY_MODULE_USAGE, " ( remote.name=<remote> ) "
+				"( node.latency=<latency as fraction> ) "
+				"( audio.rate=<sample rate> ) "
+				"( audio.channels=<number of channels> ) "
+				"( audio.position=<channel map> ) "
+				"( buffer.max_size=<max buffer size in ms> ) "
+				"( buffer.play_delay=<delay as fraction> ) "
+				"( library.name =<library name> ) "
+				"( aec.args=<aec arguments> ) "
+				"( capture.props=<properties> ) "
+				"( source.props=<properties> ) "
+				"( sink.props=<properties> ) "
+				"( playback.props=<properties> ) " },
 	{ PW_KEY_MODULE_VERSION, PACKAGE_VERSION },
 };
 
@@ -177,10 +188,12 @@ struct impl {
 	struct pw_properties *capture_props;
 	struct pw_stream *capture;
 	struct spa_hook capture_listener;
+	struct spa_audio_info_raw capture_info;
 
 	struct pw_properties *source_props;
 	struct pw_stream *source;
 	struct spa_hook source_listener;
+	struct spa_audio_info_raw source_info;
 
 	void *rec_buffer[SPA_AUDIO_MAX_CHANNELS];
 	uint32_t rec_ringsize;
@@ -189,6 +202,7 @@ struct impl {
 	struct pw_properties *playback_props;
 	struct pw_stream *playback;
 	struct spa_hook playback_listener;
+	struct spa_audio_info_raw playback_info;
 
 	struct pw_properties *sink_props;
 	struct pw_stream *sink;
@@ -197,6 +211,7 @@ struct impl {
 	uint32_t play_ringsize;
 	struct spa_ringbuffer play_ring;
 	struct spa_ringbuffer play_delayed_ring;
+	struct spa_audio_info_raw sink_info;
 
 	void *out_buffer[SPA_AUDIO_MAX_CHANNELS];
 	uint32_t out_ringsize;
@@ -218,8 +233,50 @@ struct impl {
 	struct spa_plugin_loader *loader;
 
 	bool monitor_mode;
+
+	char wav_path[512];
+	struct wav_file *wav_file;
 };
 
+static inline void aec_run(struct impl *impl, const float *rec[], const float *play[],
+		float *out[], uint32_t n_samples)
+{
+	spa_audio_aec_run(impl->aec, rec, play, out, n_samples);
+
+	if (SPA_UNLIKELY(impl->wav_path[0])) {
+		if (impl->wav_file == NULL) {
+			struct wav_file_info info;
+
+			info.info.info.raw = impl->info;
+			info.info.info.raw.channels *= 3;
+			info.info.media_type = SPA_MEDIA_TYPE_audio;
+			info.info.media_subtype = SPA_MEDIA_SUBTYPE_raw;
+
+			impl->wav_file = wav_file_open(impl->wav_path,
+					"w", &info);
+			if (impl->wav_file == NULL)
+				pw_log_warn("can't open wav path '%s': %m",
+						impl->wav_path);
+		}
+		if (impl->wav_file) {
+			uint32_t i, c = impl->info.channels;
+			const float *data[c * 3];
+
+			for (i = 0; i < c; i++) {
+				data[i      ] = play[i];
+				data[i +   c] = rec[i];
+				data[i + 2*c] = out[i];
+			}
+			wav_file_write(impl->wav_file, (void*)data, n_samples);
+		} else {
+			spa_zero(impl->wav_path);
+		}
+	} else if (impl->wav_file != NULL) {
+		wav_file_close(impl->wav_file);
+		impl->wav_file = NULL;
+	}
+}
+
 static void process(struct impl *impl)
 {
 	struct pw_buffer *cout;
@@ -313,11 +370,11 @@ static void process(struct impl *impl)
 				pd[i] = play_delayed[i] + delay_left;
 				o[i] = out[i] + delay_left;
 			}
-			spa_audio_aec_run(impl->aec, rec, pd, o, size / sizeof(float) - delay_left);
+			aec_run(impl, rec, pd, o, size / sizeof(float) - delay_left);
 		}
 	} else {
 		/* run the canceller */
-		spa_audio_aec_run(impl->aec, rec, play_delayed, out, size / sizeof(float));
+		aec_run(impl, rec, play_delayed, out, size / sizeof(float));
 	}
 
 	/* Next, copy over the output to the output ringbuffer */
@@ -507,6 +564,27 @@ static void source_state_changed(void *data, enum pw_stream_state old,
 	}
 }
 
+static void reset_buffers(struct impl *impl)
+{
+	uint32_t index, i;
+
+	spa_ringbuffer_init(&impl->rec_ring);
+	spa_ringbuffer_init(&impl->play_ring);
+	spa_ringbuffer_init(&impl->play_delayed_ring);
+	spa_ringbuffer_init(&impl->out_ring);
+
+	for (i = 0; i < impl->info.channels; i++) {
+		memset(impl->rec_buffer[i], 0, impl->rec_ringsize);
+		memset(impl->play_buffer[i], 0, impl->play_ringsize);
+		memset(impl->out_buffer[i], 0, impl->out_ringsize);
+	}
+
+	spa_ringbuffer_get_write_index(&impl->play_ring, &index);
+	spa_ringbuffer_write_update(&impl->play_ring, index + (sizeof(float) * (impl->buffer_delay)));
+	spa_ringbuffer_get_read_index(&impl->play_ring, &index);
+	spa_ringbuffer_read_update(&impl->play_ring, index + (sizeof(float) * (impl->buffer_delay)));
+}
+
 static void input_param_latency_changed(struct impl *impl, const struct spa_pod *param)
 {
 	struct spa_latency_info latency;
@@ -525,21 +603,64 @@ static void input_param_latency_changed(struct impl *impl, const struct spa_pod
 	else
 		pw_stream_update_params(impl->capture, params, 1);
 }
+
 static struct spa_pod* get_props_param(struct impl* impl, struct spa_pod_builder* b)
 {
-	if (spa_audio_aec_get_params(impl->aec, NULL) > 0) {
-		struct spa_pod_frame f[2];
-		spa_pod_builder_push_object(
-		    b, &f[0], SPA_TYPE_OBJECT_Props, SPA_PARAM_Props);
-		spa_pod_builder_prop(b, SPA_PROP_params, 0);
-		spa_pod_builder_push_struct(b, &f[1]);
+	struct spa_pod_frame f[2];
 
+	spa_pod_builder_push_object(
+	    b, &f[0], SPA_TYPE_OBJECT_Props, SPA_PARAM_Props);
+	spa_pod_builder_prop(b, SPA_PROP_params, 0);
+	spa_pod_builder_push_struct(b, &f[1]);
+
+	spa_pod_builder_string(b, "debug.aec.wav-path");
+	spa_pod_builder_string(b, impl->wav_path);
+
+	if (spa_audio_aec_get_params(impl->aec, NULL) > 0)
 		spa_audio_aec_get_params(impl->aec, b);
 
-		spa_pod_builder_pop(b, &f[1]);
-		return spa_pod_builder_pop(b, &f[0]);
+	spa_pod_builder_pop(b, &f[1]);
+	return spa_pod_builder_pop(b, &f[0]);
+}
+
+static int set_params(struct impl* impl, const struct spa_pod *params)
+{
+	struct spa_pod_parser prs;
+	struct spa_pod_frame f;
+	int changed = 0;
+
+	spa_pod_parser_pod(&prs, params);
+	if (spa_pod_parser_push_struct(&prs, &f) < 0)
+		return 0;
+
+	while (true) {
+		const char *name;
+		struct spa_pod *pod;
+		char value[512];
+
+		if (spa_pod_parser_get_string(&prs, &name) < 0)
+			break;
+
+		if (spa_pod_parser_get_pod(&prs, &pod) < 0)
+			break;
+
+		if (spa_pod_is_string(pod)) {
+			spa_pod_copy_string(pod, sizeof(value), value);
+		} else if (spa_pod_is_none(pod)) {
+			spa_zero(value);
+		} else
+			continue;
+
+		pw_log_info("key:'%s' val:'%s'", name, value);
+
+		if (spa_streq(name, "debug.aec.wav-path")) {
+			spa_scnprintf(impl->wav_path,
+				sizeof(impl->wav_path), "%s", value);
+			changed++;
+		}
 	}
-	return NULL;
+	spa_audio_aec_set_params(impl->aec, params);
+	return 1;
 }
 
 static void input_param_changed(void *data, uint32_t id, const struct spa_pod* param)
@@ -547,7 +668,12 @@ static void input_param_changed(void *data, uint32_t id, const struct spa_pod* p
 	struct spa_pod_object* obj = (struct spa_pod_object*)param;
 	const struct spa_pod_prop* prop;
 	struct impl* impl = data;
+
 	switch (id) {
+	case SPA_PARAM_Format:
+		if (param == NULL)
+			reset_buffers(impl);
+		break;
 	case SPA_PARAM_Latency:
 		input_param_latency_changed(impl, param);
 		break;
@@ -556,18 +682,18 @@ static void input_param_changed(void *data, uint32_t id, const struct spa_pod* p
 			uint8_t buffer[1024];
 			struct spa_pod_dynamic_builder b;
 			const struct spa_pod* params[1];
-			SPA_POD_OBJECT_FOREACH(obj, prop)
-			{
-				if (prop->key == SPA_PROP_params) {
-					spa_audio_aec_set_params(impl->aec, &prop->value);
-				}
+
+			SPA_POD_OBJECT_FOREACH(obj, prop) {
+				if (prop->key == SPA_PROP_params)
+					set_params(impl, &prop->value);
 			}
 
 			spa_pod_dynamic_builder_init(&b, buffer, sizeof(buffer), 4096);
 			params[0] = get_props_param(impl, &b.b);
 			if (params[0]) {
 				pw_stream_update_params(impl->capture, params, 1);
-				pw_stream_update_params(impl->playback, params, 1);
+				if (impl->playback != NULL)
+					pw_stream_update_params(impl->playback, params, 1);
 			}
 			spa_pod_dynamic_builder_clean(&b);
 		} else {
@@ -648,7 +774,12 @@ static void output_param_changed(void *data, uint32_t id, const struct spa_pod *
 	struct spa_pod_object *obj = (struct spa_pod_object *) param;
 	const struct spa_pod_prop *prop;
 	struct impl *impl = data;
+
 	switch (id) {
+	case SPA_PARAM_Format:
+		if (param == NULL)
+			reset_buffers(impl);
+		break;
 	case SPA_PARAM_Latency:
 		output_param_latency_changed(impl, param);
 		break;
@@ -668,7 +799,8 @@ static void output_param_changed(void *data, uint32_t id, const struct spa_pod *
 			params[0] = get_props_param(impl, &b.b);
 			if (params[0] != NULL) {
 				pw_stream_update_params(impl->capture, params, 1);
-				pw_stream_update_params(impl->playback, params, 1);
+				if (impl->playback != NULL)
+					pw_stream_update_params(impl->playback, params, 1);
 			}
 			spa_pod_dynamic_builder_clean(&b);
 		}
@@ -776,7 +908,6 @@ static int setup_streams(struct impl *impl)
 	uint32_t offsets[512];
 	const struct spa_pod *params[512];
 	struct spa_pod_dynamic_builder b;
-	uint32_t index;
 
 	impl->capture = pw_stream_new(impl->core,
 			"Echo-Cancel Capture", impl->capture_props);
@@ -826,7 +957,7 @@ static int setup_streams(struct impl *impl)
 	spa_pod_dynamic_builder_init(&b, NULL, 0, 4096);
 
 	offsets[n_params++] = b.b.state.offset;
-	spa_format_audio_raw_build(&b.b, SPA_PARAM_EnumFormat, &impl->info);
+	spa_format_audio_raw_build(&b.b, SPA_PARAM_EnumFormat, &impl->capture_info);
 
 	int nbr_of_external_props = spa_audio_aec_enum_props(impl->aec, 0, NULL);
 	if (nbr_of_external_props > 0) {
@@ -837,9 +968,8 @@ static int setup_streams(struct impl *impl)
 		get_props_param(impl, &b.b);
 	}
 
-	for (i = 0; i < n_params; i++) {
+	for (i = 0; i < n_params; i++)
 		params[i] = spa_pod_builder_deref(&b.b, offsets[i]);
-	}
 
 	if ((res = pw_stream_connect(impl->capture,
 			PW_DIRECTION_INPUT,
@@ -852,6 +982,12 @@ static int setup_streams(struct impl *impl)
 		return res;
 	}
 
+	offsets[0] = b.b.state.offset;
+	spa_format_audio_raw_build(&b.b, SPA_PARAM_EnumFormat, &impl->source_info);
+
+	for (i = 0; i < n_params; i++)
+		params[i] = spa_pod_builder_deref(&b.b, offsets[i]);
+
 	if ((res = pw_stream_connect(impl->source,
 			PW_DIRECTION_OUTPUT,
 			PW_ID_ANY,
@@ -862,6 +998,12 @@ static int setup_streams(struct impl *impl)
 		return res;
 	}
 
+	offsets[0] = b.b.state.offset;
+	spa_format_audio_raw_build(&b.b, SPA_PARAM_EnumFormat, &impl->sink_info);
+
+	for (i = 0; i < n_params; i++)
+		params[i] = spa_pod_builder_deref(&b.b, offsets[i]);
+
 	if ((res = pw_stream_connect(impl->sink,
 			PW_DIRECTION_INPUT,
 			PW_ID_ANY,
@@ -873,6 +1015,12 @@ static int setup_streams(struct impl *impl)
 		return res;
 	}
 
+	offsets[0] = b.b.state.offset;
+	spa_format_audio_raw_build(&b.b, SPA_PARAM_EnumFormat, &impl->playback_info);
+
+	for (i = 0; i < n_params; i++)
+		params[i] = spa_pod_builder_deref(&b.b, offsets[i]);
+
 	if (impl->playback != NULL && (res = pw_stream_connect(impl->playback,
 			PW_DIRECTION_OUTPUT,
 			PW_ID_ANY,
@@ -894,15 +1042,8 @@ static int setup_streams(struct impl *impl)
 		impl->play_buffer[i] = malloc(impl->play_ringsize);
 		impl->out_buffer[i] = malloc(impl->out_ringsize);
 	}
-	spa_ringbuffer_init(&impl->rec_ring);
-	spa_ringbuffer_init(&impl->play_ring);
-	spa_ringbuffer_init(&impl->play_delayed_ring);
-	spa_ringbuffer_init(&impl->out_ring);
 
-	spa_ringbuffer_get_write_index(&impl->play_ring, &index);
-	spa_ringbuffer_write_update(&impl->play_ring, index + (sizeof(float) * (impl->buffer_delay)));
-	spa_ringbuffer_get_read_index(&impl->play_ring, &index);
-	spa_ringbuffer_read_update(&impl->play_ring, index + (sizeof(float) * (impl->buffer_delay)));
+	reset_buffers(impl);
 
 	return 0;
 }
@@ -1104,6 +1245,11 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 
 	parse_audio_info(props, &impl->info);
 
+	impl->capture_info = impl->info;
+	impl->source_info = impl->info;
+	impl->sink_info = impl->info;
+	impl->playback_info = impl->info;
+
 	if ((str = pw_properties_get(props, "capture.props")) != NULL)
 		pw_properties_update_string(impl->capture_props, str, strlen(str));
 	if ((str = pw_properties_get(props, "source.props")) != NULL)
@@ -1157,17 +1303,21 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	if ((path = pw_properties_get(props, "library.name")) == NULL)
 		path = "aec/libspa-aec-webrtc";
 
-	struct spa_dict_item info_items[] = {
-		{ SPA_KEY_LIBRARY_NAME, path },
-	};
-	struct spa_dict info = SPA_DICT_INIT_ARRAY(info_items);
+	const struct spa_support *support;
+	uint32_t n_support;
 
-	impl->loader = spa_support_find(context->support, context->n_support, SPA_TYPE_INTERFACE_PluginLoader);
+	support = pw_context_get_support(context, &n_support);
+	impl->loader = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_PluginLoader);
 	if (impl->loader == NULL) {
 		pw_log_error("a plugin loader is needed");
 		return -EINVAL;
 	}
 
+	struct spa_dict_item info_items[] = {
+		{ SPA_KEY_LIBRARY_NAME, path },
+	};
+	struct spa_dict info = SPA_DICT_INIT_ARRAY(info_items);
+
 	handle = spa_plugin_loader_load(impl->loader, SPA_NAME_AEC, &info);
 	if (handle == NULL) {
 		pw_log_error("aec plugin %s not available library.name %s", SPA_NAME_AEC, path);
@@ -1257,6 +1407,24 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	copy_props(impl, props, SPA_KEY_AUDIO_POSITION);
 	copy_props(impl, props, "resample.prefill");
 
+	if ((str = pw_properties_get(impl->capture_props, SPA_KEY_AUDIO_POSITION)) != NULL)
+		parse_position(&impl->capture_info, str, strlen(str));
+	if ((str = pw_properties_get(impl->source_props, SPA_KEY_AUDIO_POSITION)) != NULL)
+		parse_position(&impl->source_info, str, strlen(str));
+	if ((str = pw_properties_get(impl->sink_props, SPA_KEY_AUDIO_POSITION)) != NULL)
+		parse_position(&impl->sink_info, str, strlen(str));
+	if ((str = pw_properties_get(impl->playback_props, SPA_KEY_AUDIO_POSITION)) != NULL)
+		parse_position(&impl->playback_info, str, strlen(str));
+
+	if (impl->capture_info.channels != impl->info.channels)
+		impl->capture_info = impl->info;
+	if (impl->source_info.channels != impl->info.channels)
+		impl->source_info = impl->info;
+	if (impl->sink_info.channels != impl->info.channels)
+		impl->sink_info = impl->info;
+	if (impl->playback_info.channels != impl->info.channels)
+		impl->playback_info = impl->info;
+
 	impl->max_buffer_size = pw_properties_get_uint32(props,"buffer.max_size", MAX_BUFSIZE_MS);
 
 	if ((str = pw_properties_get(props, "buffer.play_delay")) != NULL) {
